#include "vm.h"

#define COMPARE_AND_JUMP(type, accessor, op)                                   \
  do {                                                                         \
    type value = frame->registers[src1].accessor;                              \
    type value2 = frame->registers[src2].accessor;                             \
    frame->pc = (value op value2) ? dest : frame->pc;                          \
    return frame->pc;                                                          \
  } while (0)

#define MATH_OP(accessor, op)                                                  \
  do {                                                                         \
    frame->registers[dest].accessor =                                          \
        frame->registers[src1].accessor op frame->registers[src2].accessor;    \
    return frame->pc;                                                          \
  } while (0)

/**
 * String copy in data memory.
 */
void mem_strcpy(Value *memory, const char *str, uint32_t length,
                uint32_t dest) {
  memory[dest].u = length;
  uint32_t buffer = dest + 1;
  uint32_t i;
  for (i = 0; i < length; i++) {
    memory[buffer + (i / 4)].c[i % 4] = str[i];
  }
}

/**
 * Step to the next opcode in the vm.
 */
uint32_t step_vm(VMFrame *frame, Value *memory) {
  uint32_t instruction = memory[frame->pc].u;

  /* Extract 8-bit register indices from 32-bit instruction */
  uint8_t opcode = (instruction >> 24) & 0xFF;
  uint8_t dest = (instruction >> 16) & 0xFF;
  uint8_t src1 = (instruction >> 8) & 0xFF;
  uint8_t src2 = instruction & 0xFF;
  /* Advance to next instruction */
  frame->pc++;

  switch (opcode) {
  case OP_HALT:
    return 0;
  case OP_LOADI:
    frame->registers[dest].i = memory[frame->pc++].i;
    return frame->pc;
  case OP_LOADU:
    frame->registers[dest].u = memory[frame->pc++].u;
    return frame->pc;
  case OP_LOADF:
    frame->registers[dest].f = memory[frame->pc++].f;
    return frame->pc;
  case OP_STOREI:
    memory[frame->pc++].i = frame->registers[src1].i;
    return frame->pc;
  case OP_STOREU:
    memory[frame->pc++].u = frame->registers[src1].u;
    return frame->pc;
  case OP_STOREF:
    memory[frame->pc++].f = frame->registers[src1].f;
    return frame->pc;
  case OP_ADD_INT:
    MATH_OP(i, +);
  case OP_SUB_INT:
    MATH_OP(i, -);
  case OP_MUL_INT:
    MATH_OP(i, *);
  case OP_DIV_INT:
    MATH_OP(i, /);
  case OP_ADD_UINT:
    MATH_OP(u, +);
  case OP_SUB_UINT:
    MATH_OP(u, -);
  case OP_MUL_UINT:
    MATH_OP(u, *);
  case OP_DIV_UINT:
    MATH_OP(u, /);
  case OP_ADD_REAL:
    MATH_OP(f, +);
  case OP_SUB_REAL:
    MATH_OP(f, -);
  case OP_MUL_REAL:
    MATH_OP(f, *);
  case OP_DIV_REAL:
    MATH_OP(f, /);
  case OP_REAL_TO_INT:
    frame->registers[dest].i = (int32_t)(frame->registers[src1].f);
    return frame->pc;
  case OP_INT_TO_REAL:
    frame->registers[dest].f = (float)(frame->registers[src1].i);
    return frame->pc;
  case OP_REAL_TO_UINT:
    frame->registers[dest].u = (uint32_t)(frame->registers[src1].f);
    return frame->pc;
  case OP_UINT_TO_REAL:
    frame->registers[dest].f = (float)(frame->registers[src1].u);
    return frame->pc;
  case OP_MOV:
    frame->registers[dest] = frame->registers[src1];
    return frame->pc;
  case OP_JMP:
    frame->pc = frame->registers[src1].u; /* Jump to address */
    return frame->pc;
  case OP_JEQ_UINT: {
    COMPARE_AND_JUMP(uint32_t, u, ==);
  }
  case OP_JGT_UINT: {
    COMPARE_AND_JUMP(uint32_t, u, >);
  }
  case OP_JLT_UINT: {
    COMPARE_AND_JUMP(uint32_t, u, <);
  }
  case OP_JLE_UINT: {
    COMPARE_AND_JUMP(uint32_t, u, <=);
  }
  case OP_JGE_UINT: {
    COMPARE_AND_JUMP(uint32_t, u, >=);
  }
  case OP_JEQ_INT: {
    COMPARE_AND_JUMP(int32_t, i, ==);
  }
  case OP_JGT_INT: {
    COMPARE_AND_JUMP(int32_t, i, >);
  }
  case OP_JLT_INT: {
    COMPARE_AND_JUMP(int32_t, i, <);
  }
  case OP_JLE_INT: {
    COMPARE_AND_JUMP(int32_t, i, <=);
  }
  case OP_JGE_INT: {
    COMPARE_AND_JUMP(int32_t, i, >=);
  }
  case OP_JEQ_REAL: {
    COMPARE_AND_JUMP(int32_t, i, ==);
  }
  case OP_JGT_REAL: {
    COMPARE_AND_JUMP(int32_t, u, >);
  }
  case OP_JLT_REAL: {
    COMPARE_AND_JUMP(int32_t, u, <);
  }
  case OP_JGE_REAL: {
    COMPARE_AND_JUMP(int32_t, u, >=);
  }
  case OP_JLE_REAL: {
    COMPARE_AND_JUMP(int32_t, u, <=);
  }
  case OP_INT_TO_STRING: {
    int32_t a = (int32_t)frame->registers[src1].i;
    uint32_t str_dest = (uint32_t)frame->registers[dest].u;
    char buffer[32];
    int len = sprintf(buffer, "%d", a);
    mem_strcpy(memory, buffer, len, str_dest);
    return frame->pc;
  }
  case OP_UINT_TO_STRING: {
    uint32_t a = (uint32_t)frame->registers[src1].u;
    uint32_t str_dest = (uint32_t)frame->registers[dest].u;
    char buffer[32];
    int len = sprintf(buffer, "%d", a);
    mem_strcpy(memory, buffer, len, str_dest);
    return frame->pc;
  }
  case OP_REAL_TO_STRING: {
    float a = (float)frame->registers[src1].f;
    uint32_t str_dest = (uint32_t)frame->registers[dest].u;
    char buffer[32];
    int len = sprintf(buffer, "%f", a);
    mem_strcpy(memory, buffer, len, str_dest);
    return frame->pc;
  }
  case OP_READ_STRING: {
    uint32_t str_dest = (uint32_t)frame->registers[dest].u;
    uint32_t buffer = str_dest + 1;
    uint32_t length = 0;
    while (1) {
      int ch = getchar();
      if (ch == '\n' || ch == EOF) {
        memory[buffer + (length / 4)].c[length % 4] = '\0';
        break;
      }
      memory[buffer + (length / 4)].c[length % 4] = ch;
      length++;
    }
    memory[str_dest].u = length;
    return frame->pc;
  }
  case OP_PRINT_STRING: {
    uint32_t ptr = (uint32_t)frame->registers[src1].u;
    uint32_t length = memory[ptr].u;
    uint32_t str_src = ptr + 1;
    uint32_t i;
    for (i = 0; i < length; i++) {
      uint8_t ch = memory[str_src + (i / 4)].c[i % 4];
      if (ch == '\0')
        break;
      putchar(ch);
    }
    putchar('\n');
    return frame->pc;
  }
  case OP_CMP_STRING: {
    uint32_t addr1 = (uint32_t)frame->registers[src1].u;
    uint32_t addr2 = (uint32_t)frame->registers[src2].u;
    uint32_t length1 = memory[addr1 - 1].u;
    uint32_t length2 = memory[addr2 - 1].u;
    uint32_t equal = 1;

    if (length1 != length2) {
      equal = 0;
    } else {
      uint32_t i = 0;
      while (i < length1) {
        uint32_t char1 = memory[addr1 + i].u;
        uint32_t char2 = memory[addr2 + i].u;
        if (char1 != char2) {
          equal = 0;
          break;
        }
        if ((char1 & 0xFF) == '\0' && (char2 & 0xFF) == '\0')
          return frame->pc;
      }
    }
    memory[dest].u = equal;
    return frame->pc;
  }
  }
  return 0;
}
